#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#ifndef NO_BOEHM_GC
#include <gc.h>
#endif

#include "parse.h"
#include "eval.h"

list * top = NULL;
list * last = NULL;

static char * typename_table[] = {
  "SUB_TREE", "INT", "STRING", "SYMBOL",NULL
};

char * type_lookup(int type) {
  int i = 0;
  char * x = typename_table[i];
  while (type-- && x) {
    ++i;
    x = typename_table[i];
  }
  if (x) return x;
  return "NULL";
}


void encountered_eof(char * s) {
  printf("encountered EOF %s\n",s);
  exit(1);
}

void * xmalloc(size_t size) {
  void * tmp = GC_malloc(size);
  assert(tmp!=NULL);
  return tmp;
}

void * xrealloc(void * x, size_t new_size) {
  void * y = GC_realloc(x,new_size);
  assert(y!=NULL);
  return y;
}

void expand_childs(node * n, size_t count) {
  n->childs = xrealloc(n->childs,(n->child_count+=count)*sizeof(node));
}

node * new_node(int childs) {
  node * tmp = xmalloc(sizeof(node));
  tmp->child_count = childs;
  if (childs) {
    tmp->childs = xmalloc(childs*sizeof(node *));
  } else {
    tmp->childs = NULL;
  }
  return tmp;
}

void destroy_node(node * n) {
#ifdef NO_BOEHM_GC
  if (n) {
    if (n->str_data) /* free(n->str_data) */;
    if (n->symbol_data) /* free(n->symbol_data) */;
    if (n->childs) {
      int i;
      for (i=0;i<n->child_count;i++) {
        destroy_node(n->childs[i]);
      }
      /* free(n->childs) */;
    }
    /* free(n) */;
  }
#endif
}

char * parse_str(FILE * f) {
  char * x = xmalloc(1);
  size_t x_len = 1;
  int c;
  strcpy(x,"");
  c = fgetc(f);
  if (c==EOF) {
    encountered_eof("while parsing string");
  }
  while (isalpha(c)) {
    char d[2] = { c, 0 };
    x = xrealloc(x,x_len+=1);
    strcat(x,d);
    c = fgetc(f);
    if (c==EOF) {
      encountered_eof("while parsing string");
    }
  }
  ungetc(c,f);
  return x;
}

char * parse_str2(FILE * f) {
  char * x = xmalloc(1);
  size_t x_len = 1;
  int c;
  strcpy(x,"");
  c = fgetc(f);
  if (c==EOF) {
    encountered_eof("while trying to find beginning '\"'");
  }
  if (c!='"') {
    printf("invalid char %c in parse_str2\n",c);
    exit(1);
  }
  c = fgetc(f);
  if (c==EOF) {
    encountered_eof("while trying to find ending '\"'");
  }
  while (c!='"') {
    char d[2] = {c,0};
    x = xrealloc(x,x_len+=1);
    strcat(x,d);
    c = fgetc(f);
    if (c==EOF) {
      encountered_eof("while trying to find ending '\"'");
    }
  }
  return x;
}

int isws(int c) {
  return c==' ' || c=='\t' || c=='\n' || c=='\r';
}

void skip_ws(FILE * f, int ignore_eof) {
  int c;
  do {
    c = fgetc(f);
    if (c==EOF && !ignore_eof) {
      encountered_eof("while trying to skip whitespaces");
    }
  } while (isws(c));
  ungetc(c,f);
}

node * do_parse(FILE * f) {
  skip_ws(f,1);
  return parse_subtree(f);
}

node * parse_subtree(FILE * f) {
  int c;
  node * n = new_node(0);
  c = fgetc(f);
  if (c==EOF) { /* when we encounter EOF at the beginning, we just return NULL */
    /* free(n); */
    return NULL;
  }
  if (c!='(') {
    printf("unknown char %c\n",c);
    raise(SIGSEGV);
    exit(1);
  }
  skip_ws(f,0);
  c = fgetc(f);
  if (c==EOF) {
    encountered_eof("while trying to find ')'");
  }
  while (c!=')') {
    if (c=='(') {
      ungetc(c,f);
      expand_childs(n,1);
      n->childs[n->child_count-1] = parse_subtree(f);
    } else if (isdigit(c)) {
      ungetc(c,f);
      expand_childs(n,1);
      n->childs[n->child_count-1] = parse_num(f);
    } else if (c=='\'' || isalpha(c)) {
      ungetc(c,f);
      if (isalpha(c) && n->child_count==0) {
        n->symbol_data = parse_str(f);
      } else {
        expand_childs(n,1);
        n->childs[n->child_count-1] = parse_symbol(f);
      }
    } else if (c=='"') {
      ungetc(c,f);
      expand_childs(n,1);
      n->childs[n->child_count-1] = parse_string(f);
    } else {
      printf("invalid char %c\n",c);
      exit(1);
    }
    skip_ws(f,0);
    c = fgetc(f);
    if (c==EOF) {
      encountered_eof("while trying to find ')'");
    }
  }
  n->type = SUB_TREE;
  return n;
}

node * parse_num(FILE * f) {
  node * n = new_node(0);
  unsigned int x = 0;
  int c;
  n->type = INT;
  c = fgetc(f);
  if (c==EOF) {
    encountered_eof("while trying to parse number");
  }
  while (isdigit(c)) {
    x = x*10 + c - '0';
    c = fgetc(f);
    if (c==EOF) {
      encountered_eof("while trying to parse number");
    }
  }
  n->int_data = x;
  ungetc(c,f);
  return n;
}

node * parse_symbol(FILE * f) {
  node * n = new_node(0);
  int c;
  c = fgetc(f);
  if (c==EOF) {
    encountered_eof("while trying to parse symbol");
  }
  if (c!='\'') {
    printf("invalid char %c in parse_symbol\n",c);
    exit(1);
  }
  n->type = SYMBOL;
  n->symbol_data = parse_str(f);
  return n;
}

node * parse_string(FILE * f) {
  node * n = new_node(0);
  int c;
  c = fgetc(f);
  if (c==EOF) {
    encountered_eof("while trying to parse string");
  }
  if (c!='"') {
    printf("invalid char %c in parse_string\n",c);
    exit(1);
  }
  ungetc(c,f);
  n->type = STRING;
  n->str_data = parse_str2(f);
  return n;
}

list * new_list_element(void) {
  return xmalloc(sizeof(list));
}

void add_to_list(list ** top, list ** last, node * n) {
  list * tmp;
  if (n) {
    tmp = new_list_element();
    tmp->next = NULL;
    tmp->n = n;
    if (*top==NULL) {
      *top = *last = tmp;
    } else {
      (*last)->next = tmp;
      *last = (*last)->next;
    }
  }
}

void print_subtree(node * n) {
  if (n) {
    switch (n->type) {
    case SUB_TREE:
      {
        int i;
        printf("(");
        if (n->symbol_data) {
          printf("%s ",n->symbol_data);
        }
        for (i=0;i<n->child_count;i++) {
          print_subtree(n->childs[i]);
        }
        printf(")\n");
      }
      break;
    case INT:
      printf("%u ",n->int_data);
      break;
    case STRING:
      printf("\"%s\" ",n->str_data);
      break;
    case SYMBOL:
      printf("'%s ",n->symbol_data);
      break;
    default:
      printf("unknown type %d\n",n->type);
      exit(1);
    }
  }
}

void print_parse_tree(list * top) {
  list * x;
  for (x=top;x!=NULL;x=x->next) {
    print_subtree(x->n);
  }
}

void print_result(node * r) {
  if (r) {
    switch (r->type) {
    case SUB_TREE:
      puts("<subtree>");
      break;
    case INT:
      printf("<int>%d\n",r->int_data);
      break;
    case STRING:
      printf("<string>\"%s\"\n",r->str_data);
      break;
    case SYMBOL:
      printf("<symbol>'%s\n",r->symbol_data);
      break;
    default:
      printf("unknown type %d\n",r->type);
      break;
    }
  } else {
    puts("NULL");
  }
}

void load_and_run(char * file, FILE * fptr) {
  int eof_reached = 0;
  FILE * f;
  list * my_top = NULL, * my_last = NULL;
  if (file==NULL) {
    f = fptr;
  } else {
    f = fopen(file,"r");
  }
  if (f) {
    do {
      node * n = do_parse(f);
      if (n) {
        add_to_list(&my_top,&my_last,n);
      } else {
        eof_reached = 1;
      }
    } while (!eof_reached);
    eof_reached = 0;
    {
      list * x;
      for (x=my_top;x!=NULL;x=x->next) {
        eval_subtree(x->n);
        /* print_result(result); */
      }
    }
    if (file) fclose(f);
  }
}

void print_preamble(void) {
  printf("This is scheme0\n(c) 2003 Andreas Krennmair\nLicensed under the GNU GPL\n");
}

int main(int argc, char * argv[]) {
  init_eval_system();
  load_and_run("test/init.nfl",NULL);
  /* load init file */

  if (argc<2) {
    print_preamble();
    while (!feof(stdin)) {
      node * n;
      printf("scheme0> ");
      n = do_parse(stdin);
      if (n) {
        eval_subtree(n);
      }
    }
    puts("");
    /* load_and_run(NULL,stdin); */
  } else {
    int i;
    for (i=1;i<argc;i++) {
      load_and_run(argv[1],NULL);
    }
  }
  return 0;
}
